/*
 * Copyright (c) 2011-2015 Marcus Wichelmann (http://launchpad.net/~l-admin-3)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

public class Home.Backend.HomeClient : Object {
    public signal void light_changed (string light_name, bool light_state);
    public signal void shutter_changed (string shutter_name, string shutter_state);
    public signal void switch_changed (string switch_name, bool edge);

    public bool connected { get; private set; }

    private Soup.WebsocketConnection connection;

    private string server_path;
    private string device_name;
    private string device_key;

    public HomeClient (string host, uint port, string device_name, string device_key) {
        this.server_path = "ws://%s:%u/".printf (host, port);
        this.device_name = device_name;
        this.device_key = device_key;

        reconnect ();
    }

    public void set_light (string light_name, bool light_state) {
        if (!connected) {
            start_reconnect ();

            return;
        }

        Json.Object data = new Json.Object ();
        data.set_string_member ("name", light_name);
        data.set_string_member ("state", light_state ? "on" : "off");

        send_message ("set-light", data);
    }

    public void move_shutter (string shutter_name, string direction) {
        if (!connected) {
            start_reconnect ();

            return;
        }

        Json.Object data = new Json.Object ();
        data.set_string_member ("name", shutter_name);
        data.set_string_member ("direction", direction);

        send_message ("move-shutter", data);
    }

    public void stop_shutter (string shutter_name) {
        if (!connected) {
            start_reconnect ();

            return;
        }

        Json.Object data = new Json.Object ();
        data.set_string_member ("name", shutter_name);

        send_message ("stop-shutter", data);
    }

    public void list_outputs () {
        if (!connected) {
            start_reconnect ();

            return;
        }

        send_message ("list-outputs", new Json.Object ());
    }

    private void login () {
        if (!connected) {
            start_reconnect ();

            return;
        }

        debug ("Logging in...");

        Json.Object data = new Json.Object ();
        data.set_string_member ("name", device_name);
        data.set_string_member ("key", device_key);

        send_message ("login", data);
    }

    private void send_message (string command, Json.Object data) {
        Json.Node data_node = new Json.Node (Json.NodeType.OBJECT);
        data_node.set_object (data);

        Json.Builder builder = new Json.Builder ();
        builder.begin_object ();
        builder.set_member_name ("command");
        builder.add_string_value (command);
        builder.set_member_name ("data");
        builder.add_value (data_node);
        builder.end_object ();

        Json.Generator generator = new Json.Generator ();
        generator.set_root (builder.get_root ());

        string json_request = generator.to_data (null);

        debug ("Request: %s", json_request);

        connection.send_text (json_request);
    }

    private void reconnect () {
        if (connected) {
            return;
        }

        debug ("Connecting to %s", server_path);

        Soup.Session session = new Soup.Session ();
        Soup.Message message = new Soup.Message ("GET", server_path);

        session.websocket_connect_async.begin (message, null, null, null, (obj, res) => {
            try {
                connection = session.websocket_connect_async.end (res);
                connected = true;
                connect_signals ();

                debug ("Connection to home server established.");

                login ();
            } catch (Error error) {
                critical ("Connecting to home server failed: %s", error.message);

                start_reconnect ();
            }
        });
    }

    private void start_reconnect () {
        Timeout.add (5000, () => {
            reconnect ();

            return false;
        });
    }

    private void connect_signals () {
        connection.message.connect (on_message);
        connection.error.connect (on_error);
        connection.closed.connect (on_closed);
    }

    private void disconnect_signals () {
        connection.message.disconnect (on_message);
        connection.error.disconnect (on_error);
        connection.closed.disconnect (on_closed);
    }

    private void on_message (int type, Bytes message) {
        string json_response = (string)message.get_data ();

        debug ("Response: %s", json_response);

        parse_response (json_response);
    }

    private void on_error (Error error) {
        warning ("Protocol error: %s", error.message);
    }

    private void on_closed () {
        warning ("Connection to home server closed.");

        disconnect_signals ();
        connected = false;

        start_reconnect ();
    }

    private void parse_response (string json_response) {
        try {
            Json.Parser parser = new Json.Parser ();
            parser.load_from_data (json_response);

            Json.Object root_object = parser.get_root ().get_object ();

            string command = root_object.get_string_member ("command");
            Json.Object data = root_object.get_object_member ("data");

            switch (command) {
                case "logged-in":
                    debug ("Logged in.");

                    list_outputs ();

                    break;

                case "output-list":
                    string[] light_entries = data.get_string_member ("lights").split (";");
                    string[] shutter_entries = data.get_string_member ("shutters").split (";");

                    foreach (string light_entry in light_entries) {
                        string[] fields = light_entry.split (",");

                        light_changed (fields[0], fields[1] == "on");
                    }

                    foreach (string shutter_entry in shutter_entries) {
                        string[] fields = shutter_entry.split (",");

                        shutter_changed (fields[0], fields[1]);
                    }

                    break;

                case "input-changed":
                    string input_name = data.get_string_member ("name");
                    string input_type = data.get_string_member ("type");
                    bool input_edge = data.get_string_member ("edge") == "rising";

                    debug ("%s ege on input \"%s\".", input_edge ? "Rising" : "Falling", input_name);

                    switch (input_type) {
                        case "Switch":
                            switch_changed (input_name, input_edge);

                            break;
                    }

                    break;

                case "output-changed":
                    string output_name = data.get_string_member ("name");
                    string output_type = data.get_string_member ("type");
                    string output_state = data.get_string_member ("state");

                    debug ("Output \"%s\" changed to \"%s\".", output_name, output_state);

                    switch (output_type) {
                        case "Light":
                            light_changed (output_name, output_state == "on");

                            break;
                        case "Shutter":
                            shutter_changed (output_name, output_state);

                            break;
                    }

                    break;
            }
        } catch (Error error) {
            warning ("Parsing response failed: %s", error.message);
        }
    }
}