/*
 * Copyright (c) 2011-2015 Marcus Wichelmann (http://launchpad.net/~l-admin-3)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

public class Home.Indicator : Wingpanel.Indicator {
    private Wingpanel.Widgets.OverlayIcon overlay_icon;
    private Gtk.Box main_box;

    private Backend.HomeClient? home_client = null;

    private string[] lights;
    private string[] shutters;

    private bool main_light_state;

    private Gee.HashMap<string, Wingpanel.Widgets.Switch> light_switches;
    private Gee.HashMap<string, Widgets.ShutterButton> shutter_buttons;

    public Indicator () {
        Object (code_name : "home",
                display_name: _("Home Automation"),
                description: _("The interface to your home"));

        Notify.init ("wingpanel-indicator-home");
    }

    public override Gtk.Widget get_display_widget () {
        if (overlay_icon == null) {
            overlay_icon = new Wingpanel.Widgets.OverlayIcon ("light-off-symbolic");
            overlay_icon.button_press_event.connect ((e) => {
                if (e.button == Gdk.BUTTON_MIDDLE && home_client != null) {
                    home_client.set_light (lights[0], !main_light_state);

                    return Gdk.EVENT_STOP;
                }

                return Gdk.EVENT_PROPAGATE;
            });
        }

        return overlay_icon;
    }

    public override Gtk.Widget? get_widget () {
        if (main_box == null) {
            main_box = new Gtk.Box (Gtk.Orientation.VERTICAL, 0);

            try {
                string config_file_name = "%s/.home.conf".printf (Environment.get_home_dir ());

                debug ("Reading config file %s", config_file_name);

                KeyFile config_file = new KeyFile ();
                config_file.set_list_separator (',');
                config_file.load_from_file (config_file_name, KeyFileFlags.NONE);

                home_client = new Backend.HomeClient (config_file.get_string ("Server", "host"),
                                                      (uint)config_file.get_uint64 ("Server", "port"),
                                                      config_file.get_string ("Login", "device_name"),
                                                      config_file.get_string ("Login", "device_key"));

                lights = config_file.get_string_list ("Outputs", "lights");
                shutters = config_file.get_string_list ("Outputs", "shutters");

                light_switches = new Gee.HashMap<string, Wingpanel.Widgets.Switch> ();
                shutter_buttons = new Gee.HashMap<string, Widgets.ShutterButton> ();

                foreach (string light_name in lights) {
                    Wingpanel.Widgets.Switch light_switch = new Wingpanel.Widgets.Switch (light_name);
                    light_switch.switched.connect (() => {
                        home_client.set_light (light_name, light_switch.get_active ());
                    });

                    main_box.add (light_switch);

                    light_switches.set (light_name, light_switch);
                }

                main_box.add (new Wingpanel.Widgets.Separator ());

                foreach (string shutter_name in shutters) {
                    Gtk.Grid content_widget = new Gtk.Grid ();
                    content_widget.hexpand = true;
                    content_widget.margin_start = 8;
                    content_widget.margin_end = 10;
                    content_widget.margin_top = 4;
                    content_widget.margin_bottom = 4;

                    Gtk.Label shutter_label = new Gtk.Label (shutter_name);
                    shutter_label.halign = Gtk.Align.START;
                    shutter_label.margin_start = 6;
                    shutter_label.margin_end = 10;

                    Widgets.ShutterButton shutter_button = new Widgets.ShutterButton ();
                    shutter_button.halign = Gtk.Align.END;
                    shutter_button.margin_end = 12;
                    shutter_button.move.connect ((direction) => {
                        home_client.move_shutter (shutter_name, direction);
                    });
                    shutter_button.stop.connect (() => {
                        home_client.stop_shutter (shutter_name);
                    });

                    content_widget.attach (shutter_label, 0, 0, 1, 1);
                    content_widget.attach (shutter_button, 1, 0, 1, 1);

                    main_box.add (content_widget);

                    shutter_buttons.set (shutter_name, shutter_button);
                }

                connect_signals ();
            } catch (Error e) {
                critical ("Reading config file failed: %s", e.message);
            }
        }

        return main_box;
    }

    public override void opened () {
    }

    public override void closed () {
    }

    private void connect_signals () {
        home_client.notify["connected"].connect ((state) => {
            this.visible = home_client.connected;
        });

        home_client.light_changed.connect ((light_name, light_state) => {
            Wingpanel.Widgets.Switch? light_switch = light_switches.get (light_name);

            if (light_switch != null) {
                light_switch.set_active (light_state);
            }

            if (light_name == lights[0]) {
                main_light_state = light_state;

                overlay_icon.set_main_icon_name (light_state ? "light-on-symbolic" : "light-off-symbolic");
            }
        });

        home_client.shutter_changed.connect ((shutter_name, shutter_state) => {
            Widgets.ShutterButton? shutter_button = shutter_buttons.get (shutter_name);

            if (shutter_button != null) {
                shutter_button.set_visible_child_name (shutter_state);
            }
        });

        home_client.switch_changed.connect ((switch_name, edge) => {
            if (switch_name == "Marcus Tür 2A" && edge) {
                toggle_music ();
            }
        });
    }

    private void toggle_music () {
        Bus.get.begin (BusType.SESSION, null, (obj, res) => {
            try {
                DBusConnection bus = Bus.get.end (res);
                bus.call.begin ("org.pantheon.noise",
                                "/org/mpris/MediaPlayer2",
                                "org.mpris.MediaPlayer2.Player",
                                "PlayPause",
                                null,
                                null,
                                DBusCallFlags.NONE,
                                -1,
                                null, (obj, res) => {
                    try {
                        bus.call.end (res);
                    } catch (Error error) {
                        warning ("Triggering music failed: %s", error.message);
                    }
                });
            } catch (Error error) {
                warning ("Conencting to DBus failed: %s", error.message);
            }
        });
    }
}

public Wingpanel.Indicator? get_indicator (Module module, Wingpanel.IndicatorManager.ServerType server_type) {
    debug ("Activating Home Indicator");

    if (server_type != Wingpanel.IndicatorManager.ServerType.SESSION) {
        return null;
    }

    var indicator = new Home.Indicator ();

    return indicator;
}